import pydantic as pyd
import typing as tp
import typing_extensions as tpe

from nexus.server.core import schemas

__all__ = [
    "SingleFieldResponse",
    "JobRequest",
    "JobUpdateRequest",
    "JobListRequest",
    "JobLogsResponse",
    "ServerLogsResponse",
    "ServerActionResponse",
    "GpuActionResponse",
    "GpuStatusResponse",
    "ServerStatusResponse",
    "HealthResponse",
    "ArtifactUploadResponse",
]

REQUIRED_ENV_VARS = {
    "wandb": ["WANDB_API_KEY", "WANDB_ENTITY"],
    "nullpointer": [],
    "discord": ["DISCORD_USER_ID", "DISCORD_WEBHOOK_URL"],
    "phone": ["TWILIO_ACCOUNT_SID", "TWILIO_AUTH_TOKEN", "TWILIO_FROM_NUMBER", "PHONE_TO_NUMBER"],
}


class FrozenBaseModel(pyd.BaseModel):
    model_config = pyd.ConfigDict(frozen=True)


T = tp.TypeVar("T")


class SingleFieldResponse(FrozenBaseModel, tp.Generic[T]):
    data: T


JobLogsResponse = SingleFieldResponse[str]
ServerLogsResponse = SingleFieldResponse[str]
ServerActionResponse = SingleFieldResponse[str]
ArtifactUploadResponse = SingleFieldResponse[str]


def _check_required_vars(kinds: tp.Sequence[str], env: dict[str, str], for_type: str = "") -> None:
    for kind in kinds:
        for key in REQUIRED_ENV_VARS[kind]:
            if key not in env:
                suffix = f" for {kind} {for_type}" if for_type else f" for {kind}"
                raise ValueError(f"Missing required environment variable {key}{suffix}")


class JobRequest(FrozenBaseModel):
    job_id: str | None = None
    artifact_id: str
    command: str
    user: str
    num_gpus: int = 1
    gpu_idxs: list[int] | None = None
    priority: int = 0
    integrations: list[schemas.IntegrationType] = []
    notifications: list[schemas.NotificationType] = []
    env: dict[str, str] = {}
    jobrc: str | None = None
    run_immediately: bool = False
    ignore_blacklist: bool = False
    git_repo_url: str | None = None
    git_branch: str | None = None
    git_tag: str | None = None

    @pyd.model_validator(mode="after")
    def check_requirements(self) -> tpe.Self:
        _check_required_vars(self.integrations, self.env, "integration")
        _check_required_vars(self.notifications, self.env, "notifications")
        return self


class GpuActionError(FrozenBaseModel):
    index: int
    error: str


class GpuActionResponse(FrozenBaseModel):
    blacklisted: list[int] | None = None
    removed: list[int] | None = None
    failed: list[GpuActionError]


class GpuStatusResponse(FrozenBaseModel):
    gpu_idx: int
    blacklisted: bool
    changed: bool


class ServerStatusResponse(FrozenBaseModel):
    gpu_count: int
    queued_jobs: int
    running_jobs: int
    completed_jobs: int
    server_user: str
    server_version: str


class DiskStatsResponse(FrozenBaseModel):
    total: int
    used: int
    free: int
    percent_used: float


class NetworkStatsResponse(FrozenBaseModel):
    download_speed: float
    upload_speed: float
    ping: float


class SystemStatsResponse(FrozenBaseModel):
    cpu_percent: float
    memory_percent: float
    uptime: float
    load_avg: list[float]


class HealthResponse(FrozenBaseModel):
    alive: bool = True
    status: str | None = None
    score: float | None = None
    disk: DiskStatsResponse | None = None
    network: NetworkStatsResponse | None = None
    system: SystemStatsResponse | None = None


class JobUpdateRequest(FrozenBaseModel):
    command: str | None = None
    priority: int | None = None
    num_gpus: int | None = None
    git_tag: str | None = None


class JobListRequest(FrozenBaseModel):
    status: schemas.JobStatus | None = None
    gpu_index: int | None = None
    command_regex: str | None = None
    limit: int = 100
    offset: int = 0



import sys
import shutil

if not sys.platform.startswith("linux"):
    raise RuntimeError("This package is only compatible with the Linux operating systems")

for cmd in ("screen", "script"):
    if shutil.which(cmd) is None:
        raise RuntimeError(f"Required command '{cmd}' not found in PATH. Please install it to proceed.")



import dataclasses as dc
import pathlib as pl
import typing as tp

__all__ = ["JobStatus", "NotificationType", "IntegrationType", "Job"]


def _exclude_env_repr(obj):
    return {k: v for k, v in dc.asdict(obj).items() if k != "env"}


JobStatus = tp.Literal["queued", "running", "completed", "failed", "killed"]
NotificationType = tp.Literal["discord", "phone"]
IntegrationType = tp.Literal["wandb", "nullpointer"]


@dc.dataclass(frozen=True, slots=True)
class Job:
    id: str
    command: str
    artifact_id: str
    git_repo_url: str | None
    git_branch: str | None
    git_tag: str | None
    status: JobStatus
    created_at: float
    priority: int
    num_gpus: int
    env: dict[str, str]
    node_name: str
    jobrc: str | None
    integrations: list[IntegrationType]
    notifications: list[NotificationType]
    notification_messages: dict[str, str]
    pid: int | None
    dir: pl.Path | None
    started_at: float | None
    gpu_idxs: list[int]
    wandb_url: str | None
    marked_for_kill: bool
    completed_at: float | None
    exit_code: int | None
    error_message: str | None
    user: str
    ignore_blacklist: bool
    screen_session_name: str | None

    def __repr__(self) -> str:
        return str(_exclude_env_repr(self))



import aiohttp

from nexus.server.core import exceptions as exc
from nexus.server.utils import logger


@exc.handle_exception(
    Exception, exc.NotificationError, message="0x0.st upload failed", reraise=False, default_return=None
)
async def upload_text_to_nullpointer(text: str, instance_url: str = "https://0x0.st/") -> str | None:
    if not text:
        return None

    form_data = aiohttp.FormData()
    form_data.add_field("file", text.encode("utf-8"), filename="job_logs.txt", content_type="text/plain")
    form_data.add_field("secret", "1")
    form_data.add_field("expires", "24")

    async with aiohttp.ClientSession() as session:
        async with session.post(instance_url, data=form_data) as response:
            if response.status == 200:
                url = await response.text()
                url = url.strip()
                return url
            else:
                error_text = await response.text()
                logger.error(f"0x0.st upload failed with status {response.status}: {error_text}")

    return None
